#include "Wire.h"


bool debug = true;

byte photointerrupters000B = 0; // I/O 000 B: C, C#, D, D#
byte photointerrupters000A = 0; // I/O 000 A: E, F, F#, G
byte photointerrupters001B = 0; // I/O 001 B: G#, A, A#, B

byte indicator001A = 0;
byte indicator002A = 0;
byte indicator003B = 0;

bool pressing[12] = {0,0,0,0,0,0,0,0,0,0,0,0}; // Signifies that the top photointerrupter is triggered
unsigned long pressingStartTime[12] = {0,0,0,0,0,0,0,0,0,0,0,0}; // Start time of the press
bool pressed[12] = {0,0,0,0,0,0,0,0,0,0,0,0}; // Signifies that both photointerrupters have been triggered and the note has been played.
unsigned long currentTime;

unsigned long softestPressTime = 100;
byte minimumVolume = 30;

byte COLOR_SELECT_0 = 17;
byte COLOR_SELECT_1 = 16;
byte COLOR_SELECT_2 = 15;
byte COLOR_SELECT_3 = 14;



struct NextKey {
  byte index;
  byte key;
  byte finger;
  bool played;
};

byte mode = 1;
int songIterator = 0;
int songIndex = 0;
int songLength = 42;
NextKey nextKeys[42];
bool needsUpdate = true;

//NextKey nextKey;

void setup() {
  Serial.begin(9600);
  Wire.begin();
  
  //Wire.setClock(1700000); // Set clock to 1.7Mhz
  //Wire.setClock(400000); // Set clock to 400khz
  Wire.setClock(400000);

  pinMode(COLOR_SELECT_0, OUTPUT);
  pinMode(COLOR_SELECT_1, OUTPUT);
  pinMode(COLOR_SELECT_2, OUTPUT);
  pinMode(COLOR_SELECT_3, OUTPUT);

  // Set up song - Twinkle Twinkle Little Star - CCGGAAG FFEEDDC
  nextKeys[0] = {0, 0, 5, false}; // C, R thumb
  nextKeys[1] = {1, 0, 5, false}; // C, R thumb
  nextKeys[2] = {2, 7, 8, false}; // G, R ring
  nextKeys[3] = {3, 7, 8, false}; // G, R ring
  nextKeys[4] = {4, 9, 9, false}; // A, R pinky
  nextKeys[5] = {5, 9, 9, false}; // A, R pinky
  nextKeys[6] = {6, 7, 8, false}; // G, R ring

  nextKeys[7] = {7, 5, 8, false}; // F, R ring
  nextKeys[8] = {8, 5, 8, false}; // F, R ring
  nextKeys[9] = {9, 4, 7, false}; // E, R middle
  nextKeys[10] = {10, 4, 7, false}; // E, R middle
  nextKeys[11] = {11, 2, 6, false}; // D, R index
  nextKeys[12] = {12, 2, 6, false}; // D, R index
  nextKeys[13] = {13, 0, 5, false}; // C, R thumb

  nextKeys[14] = {14, 7, 0, false};
  nextKeys[15] = {15, 7, 0, false};
  nextKeys[16] = {16, 5, 0, false};
  nextKeys[17] = {17, 5, 0, false};
  nextKeys[18] = {18, 4, 0, false};
  nextKeys[19] = {19, 4, 0, false};
  nextKeys[20] = {20, 2, 0, false};
  
  nextKeys[21] = {21, 7, 0, false};
  nextKeys[22] = {22, 7, 0, false};
  nextKeys[23] = {23, 5, 0, false};
  nextKeys[24] = {24, 5, 0, false};
  nextKeys[25] = {25, 4, 0, false};
  nextKeys[26] = {26, 4, 0, false};
  nextKeys[27] = {27, 2, 0, false};

  nextKeys[28] = {0, 0, 5, false}; // C, R thumb
  nextKeys[29] = {1, 0, 5, false}; // C, R thumb
  nextKeys[30] = {2, 7, 8, false}; // G, R ring
  nextKeys[31] = {3, 7, 8, false}; // G, R ring
  nextKeys[32] = {4, 9, 9, false}; // A, R pinky
  nextKeys[33] = {5, 9, 9, false}; // A, R pinky
  nextKeys[34] = {6, 7, 8, false}; // G, R ring

  nextKeys[35] = {7, 5, 8, false}; // F, R ring
  nextKeys[36] = {8, 5, 8, false}; // F, R ring
  nextKeys[37] = {9, 4, 7, false}; // E, R middle
  nextKeys[38] = {10, 4, 7, false}; // E, R middle
  nextKeys[39] = {11, 2, 6, false}; // D, R index
  nextKeys[40] = {12, 2, 6, false}; // D, R index
  nextKeys[41] = {13, 0, 5, false}; // C, R thumb

  // Set I/O Expander 000 Register A to input
  Wire.beginTransmission(0x20);
  Wire.write(0x00);
  Wire.write(0xFF);
  Wire.endTransmission();

  // Set I/O Expander 000 Register B to input
  Wire.beginTransmission(0x20);
  Wire.write(0x01);
  Wire.write(0xFF);
  Wire.endTransmission();

  // Set I/O Expander 001 Register B to input
  Wire.beginTransmission(0x21);
  Wire.write(0x01);
  Wire.write(0xFF);
  Wire.endTransmission();

  SetIOMode(0x21, 0x00, 0xFF); // I/O Expander 001, Register A input
  SetIOMode(0x23, 0x00, 0x00); // I/O Expander 003, Register A output
  SetIOMode(0x23, 0x01, 0x00); // I/O Expander 003, Register B output
  SetIOMode(0x24, 0x00, 0x00); // I/O Expander 004, Register A output
  SetIOMode(0x24, 0x01, 0x00); // I/O Expander 004, Register B output

  Wire.beginTransmission(0x24);
  Wire.write(0x12);
  Wire.write(255);
  Wire.endTransmission();

  Wire.beginTransmission(0x24);
  Wire.write(0x13);
  Wire.write(255);
  Wire.endTransmission();

  delay(1000);
  /*
  while (Serial.available() == 0) {}
  if (true) {
    mode = 1;
    byte input = 0;
    Serial.readBytes(input, 1);
    nextKey = {0, input, 0, false};
    needsUpdate = true;
    Serial.print("Serial Bytes Available: ");
    Serial.print(Serial.available());
    Serial.print(" Byte Read: ");
    Serial.print(input);
    Serial.print("\n");
  } else {
    mode = 0;  
    needsUpdate = true;
  }
  */
  
  Serial.write("Program initialized.\n");
}

void SetIOMode(byte address, byte reg, byte mode) {
 Wire.beginTransmission(address);
 Wire.write(reg);
 Wire.write(mode);
 Wire.endTransmission(); 
}

void loop() {
  if (mode != 0) {
    if (songIndex >= songLength) mode = 0;
    if (needsUpdate) {
      byte key = nextKeys[songIndex].key;
      //byte key = nextKey.key;

      byte indicatorA = 0;
      byte indicatorB = 0;

      if (key == 0) bitSet(indicatorB, 7);
      if (key == 1) bitSet(indicatorB, 6);
      if (key == 2) bitSet(indicatorB, 5);
      if (key == 3) bitSet(indicatorB, 4);
      if (key == 4) bitSet(indicatorB, 3);
      if (key == 5) bitSet(indicatorB, 2);
      if (key == 6) bitSet(indicatorB, 1);
      if (key == 7) bitSet(indicatorB, 0);
      if (key == 8) bitSet(indicatorA, 0);
      if (key == 9) bitSet(indicatorA, 1);
      if (key == 10) bitSet(indicatorA, 2);
      if (key == 11) bitSet(indicatorA, 3);

      Wire.beginTransmission(0x24);
      Wire.write(0x12);
      Wire.write(indicatorA);
      Wire.endTransmission();

      Wire.beginTransmission(0x24);
      Wire.write(0x13);
      Wire.write(indicatorB);
      Wire.endTransmission();
    }  
  } else if (needsUpdate) {
      Wire.beginTransmission(0x24);
      Wire.write(0x12);
      Wire.write(255);
      Wire.endTransmission();

      Wire.beginTransmission(0x24);
      Wire.write(0x13);
      Wire.write(255);
      Wire.endTransmission();
  }
  /*
  for (int i = songIterator; i < 1000; i++) {
    if (!nextKeys[i].played) break;
    else if (nextKeys[i].index != songIndex) {
      songIterator = i;
      songIndex++;
      break;
    }
  }
  */
  
  // Poll I/O Expander 000 for Register B
  Wire.beginTransmission(0x20); // 010 0000
  Wire.write(0x13); // Select Register B
  Wire.endTransmission();
  Wire.requestFrom(0x20, 1);
  photointerrupters000B = Wire.read();
  
  // Poll I/O Expander 000 for Register A
  Wire.beginTransmission(0x20); // 010 0000
  Wire.write(0x12); // Select Register A
  Wire.endTransmission();
  Wire.requestFrom(0x20, 1);
  photointerrupters000A = Wire.read();

  // Poll I/O Expander 001 for Register B
  Wire.beginTransmission(0x21); // 010 0001
  Wire.write(0x13);
  Wire.endTransmission();
  Wire.requestFrom(0x21, 1);
  photointerrupters001B = Wire.read();

  
  currentTime = millis();

  ProcessKey(!bitRead(photointerrupters000B, 0), !bitRead(photointerrupters000B, 1),  0); // C
  ProcessKey(!bitRead(photointerrupters000B, 2), !bitRead(photointerrupters000B, 3),  1); // C#
  ProcessKey(!bitRead(photointerrupters000B, 4), !bitRead(photointerrupters000B, 5),  2); // D
  ProcessKey(!bitRead(photointerrupters000B, 6), !bitRead(photointerrupters000B, 7),  3); // D#
  ProcessKey(!bitRead(photointerrupters000A, 0), !bitRead(photointerrupters000A, 1),  4); // E
  ProcessKey(!bitRead(photointerrupters000A, 2), !bitRead(photointerrupters000A, 3),  5); // F
  ProcessKey(!bitRead(photointerrupters000A, 4), !bitRead(photointerrupters000A, 5),  6); // F#
  ProcessKey(!bitRead(photointerrupters000A, 6), !bitRead(photointerrupters000A, 7),  7); // G
  
  ProcessKey(!bitRead(photointerrupters001B, 0), !bitRead(photointerrupters001B, 1),  8); // G#
  ProcessKey(!bitRead(photointerrupters001B, 2), !bitRead(photointerrupters001B, 3),  9); // A
  ProcessKey(!bitRead(photointerrupters001B, 4), !bitRead(photointerrupters001B, 5), 10); // A#
  ProcessKey(!bitRead(photointerrupters001B, 6), !bitRead(photointerrupters001B, 7), 11); // B
  
  //Serial.print(pressing[0])
  //Serial.print(" - ");
  //Serial.print(pressed[0]);
  //Serial.print("\n");
  //Serial.print(photointerrupters000B);
  //Serial.print("\n");
}

void ProcessKey(bool inputTop, bool inputBottom, byte keyIndex) {
  if (!pressed[keyIndex]) { // Key has not been pressed fully yet
    if (inputTop) { // Key is being pressed.
      if (!pressing[keyIndex]) { // Key was not being pressed before. Record pressing and time.
        //Serial.print("Pressing\n");
        pressing[keyIndex] = true;
        pressingStartTime[keyIndex] = currentTime;
      } else { // Key is waiting to be pressed fully.
        if (inputBottom) { // Key has been pressed fully. Record pressed and play the note.
          pressed[keyIndex] = true;
          unsigned long totalTime = currentTime - pressingStartTime[keyIndex];
          if (totalTime > softestPressTime) totalTime = softestPressTime; // Longest keypress is 1s.
          float v = (float) (softestPressTime - totalTime) / (float) softestPressTime; // Volume is a normalized value, 0.0 to 1.0
          byte volume = (byte) (v * (255 - minimumVolume)) + minimumVolume;
          PlayNote(volume, keyIndex);
        }
      }
    } else { // Key was released early. Reset pressing.
      pressing[keyIndex] = false;
    }
  } else { // Key is still pressed and note has been played
    if (!inputTop) { // Key has been released, reset pressing and pressed
      pressing[keyIndex] = false;
      pressed[keyIndex] = false;
    }
  }
}

void PlayNote(byte volume, byte keyIndex) {
  if (mode == -1) {
    
  }
  else if (mode == 0) {
    // Song is not loaded. Play the key with no validation
    ProcessNote(volume, keyIndex);
    
  } else if (mode == 1) {
    // Song is loaded. Play the song with key validation
    bool correctKey = ValidateKey(keyIndex);
    //if (correctKey) ProcessNote(volume, keyIndex);
    ProcessNote(volume, keyIndex);
  } else {
    // Song is loaded. Play the song with key and finger validation
    bool correctKey = ValidateKey(keyIndex);
    if (correctKey) {
      // Correct Key
      
      byte finger = PollColorSensor(keyIndex);
      if (finger == nextKeys[songIterator].finger) {
        // Correct Finger        
        ProcessNote(volume, keyIndex);
      } else {
        // Incorrect Finger
       
        // Don't send note to audio
        // Send data to computer
      }
    } else {
      // Send data to computer  
    }
  }
}

bool ValidateKey(byte keyIndex) {
  //Serial.print(songIterator);
  //Serial.print("\n");
  //Serial.print(songIndex);
  //Serial.print("\n");
  //Serial.print("\n");
  
  bool correctKey = false;
  /*
  if (mode == 1) {
    if (keyIndex == nextKey.key) {
      correctKey = true;
      needsUpdate = true;  
    }
  }
  */

  
  if (nextKeys[songIndex].key == keyIndex) {
    songIndex++;
    correctKey = true;
    needsUpdate = true;
  }  
  
  
  
  return correctKey;
}

byte PollColorSensor(byte keyIndex) {
  byte finger;

  // Set select bits
  digitalWrite(COLOR_SELECT_0, bitRead(keyIndex, 0) ? HIGH : LOW);
  digitalWrite(COLOR_SELECT_1, bitRead(keyIndex, 1) ? HIGH : LOW);
  digitalWrite(COLOR_SELECT_2, bitRead(keyIndex, 2) ? HIGH : LOW);
  digitalWrite(COLOR_SELECT_3, bitRead(keyIndex, 3) ? HIGH : LOW);

  // Poll color sensor, store it in finger

  return finger;
}

void ProcessNote(byte volume, byte note) {
    Serial.write(note);
    Serial.write(volume);
    Serial.write("\n");
    
    //byte input = Serial.read();
    //if (input == 255) mode = 0;
    //else nextKey = {0, input, 0, false};

    /*
    if (debug) {
      Serial.write("Mode: ");
      Serial.write(mode);
      Serial.write(" Played Key ");
      Serial.write(note);
      

      if (mode != 0) {
        Serial.write(", Next Key: ");
        Serial.write(nextKey.key);
      }
      Serial.write("\n");
    }
    */
    
    //Wire.beginTransmission(audioAddress);
    //Wire.write(volume);
    //Wire.write(note);
    //Wire.endTransmission();
}
